% +-------------------------------+
% | Algorithm: QRD-RLS (convent.) |
% | Jos Antonio Apolinrio Jr.   |
% +-------------------------------+
clear; clc; close all; pack;
K        = 1000;       % number of iterations
ENS      = 5;          % ensemble
varx     = 1;          % inout signal variance
varn     = 1e-8;       % noise variance
nrcoef   = 4;          % number of coeficients (N+1)
N        = nrcoef-1;   % filter order
num      = [1];        % num and den used to generate colored input signal
den      = [1 1.2 0.81];   
lambda   = 0.99;       % forgetting factor
lambda05 = sqrt(lambda);
Wopt     = [1 0.9*sqrt(-1) 0.1 0.2*sqrt(-1)]';
MSE      = zeros(K,1); % used to plot the learning curve
for ens=1:ENS,
  disp(ens);
  % Generating input and reference (desired) signals:
  vectorx = zeros(N+1,1);
  x       = randn(K,1)+sqrt(-1)*randn(K,1); 
  x       = filter(num,den,x); x = x/std(x);             x = x*sqrt(varx);
  noise   = randn(K,1);    noise = noise/std(noise); noise = noise*sqrt(varn);  
  %----------------
  % Initialization:
  %----------------  
  W       = zeros(N+1,1);
  dq2km1  = zeros(N+1,1);
  % Ukm1 may be initialized as a reversal matrix:
  delta   = 0.001;
  Ukm1    = zeros(N+1); for n=1:(N+1); Ukm1(n,N+2-n)=delta; end;
  % Or as:
  % Ukm1  = inv(1000*randn(N+1,N+1).*hankel(1000*randn(N+1,1)));
  % Running ...
  for k=1:K 
    vectorx = [x(k)
              vectorx(1:N)];
    % Obtaining the reference signal:
    % d = filter(Wopt,1,x)+n; ... avoid this command when implementing 
    %                             in complex environment 
    d(k)   = Wopt'*vectorx + noise(k);          
    xaux   = vectorx';      % Hermitian of the input signal vector
    Uaux   = lambda05*Ukm1;
    dq2aux = lambda05*dq2km1;
    daux   = conj(d(k));    % conjugate od d(k)
    gamma  = 1;
    for n=1:(N+1),
      %----------------------------------------  
      % Obtaining Q_theta(k) and updating U(k):  
      %----------------------------------------
      costheta      = abs(Uaux(N+2-n,n))/sqrt(abs(Uaux(N+2-n,n))^2+abs(xaux(n))^2);
      sintheta      = conj(xaux(n)/Uaux(N+2-n,n))*costheta;
      Uaux(N+2-n,n) = sintheta*xaux(n)+costheta*Uaux(N+2-n,n);
      xaux(n)=0;
      for m=n+1:N+1
          oldxaux       = xaux(m);
          xaux(m)       = costheta*oldxaux - conj(sintheta)*Uaux(N+2-n,m);
          Uaux(N+2-n,m) = sintheta*oldxaux + costheta*Uaux(N+2-n,m);
      end
      %--------------------
      % Obtaining gamma(k):  
      %--------------------
      gamma = gamma*costheta;
      %--------------------------------------
      % Obtaining eq1(k) and updating dq2(k):  
      %--------------------------------------
      olddaux       = daux;
      daux          = costheta*olddaux - conj(sintheta)*dq2aux(N+2-n);
      dq2aux(N+2-n) = sintheta*olddaux + costheta*dq2aux(N+2-n);
    end;
    %-------------------------------------------------------------------
    % Back-substitution if necessary (equivalent to W=inv(Uaux)*dq2aux):
    %-------------------------------------------------------------------
    W(N+1)=dq2aux(1)/Uaux(1,N+1);
    for ii=N:-1:1
        ll=N-ii+2;
        summa=0;
        for jj=(ii+1):(N+1)
            summa=summa+Uaux(ll,jj)*W(jj);
        end
        W(ii)=(dq2aux(ll)-summa)/Uaux(ll,ii);        
    end
    Ukm1   = Uaux;   % U(k)   = Uaux   now becoming U(k-1)   for the next iter.
    dq2km1 = dq2aux; % dq2(k) = dq2aux now becoming dq2(k-1) for the next iter.
    eq1k   = daux;   % eq1(k) = daux
    %----------------
    % Obtaining d(k):
    %----------------
    eapriori     = conj(eq1k)/gamma; % a priori     error (d(k)-Wnonupdated'*vectorx)
    eaposteriori = conj(eq1k)*gamma; % a posteriori error (d(k)-Wupdated'*vectorx)
    MSE(k)=MSE(k)+(abs(eapriori))^2; 
  end; % for k
end; % for ens
MSE      = MSE/ENS;
MSEdB    = 10*log10(MSE);
MSEmindB = 10*log10(varn)*ones(K,1);
%--------------------
% Simulation results:
%--------------------
xscale = 1:K;
plot(xscale,MSEdB); hold;
xlabel('k'); ylabel('E[|e(k)|^2]_{dB}');
title('Learning curve (a priori error) of the Conventional QRD-RLS Algorithm');
plot(xscale,MSEmindB,'--r')
%--------------------------------------------------------------------------
